/* GeoStats - IRC Statistical Services
** Copyright (c) 1999 Jonathan George
**  E-Mail: net@lite.net
*
** GeoStats CVS Identification
** $Id: users.c,v 1.17 2001/04/08 14:25:32 wcampbel Exp $
*/

#include <assert.h>
#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "patchlevel.h"

#ifdef FNMATCH
# include <fnmatch.h>
#endif /* FNMATCH */

/* These two vars are created in main.c */
extern const char version_date[];
extern const char version_time[];

/* These are ugly functions, gets right down to the hash tables and
** manipulation of raw data
*/

static Server *new_server(char *, char *);
Server *serverlist[S_TABLE_SIZE];

MyUser *myuhead;

static User *new_user(char *);
static void add_user_to_hash_table(char *, User *);
static void del_user_from_hash_table(User *);
static void add_server_to_hash_table(char *, Server *);
static void del_server_from_hash_table(Server *);

User *userlist[U_TABLE_SIZE];

static void add_user_to_hash_table(char *nick, User *u)
{
	u->hash = HASH((unsigned char *) nick, U_TABLE_SIZE);
	u->next = userlist[u->hash];
	userlist[u->hash] = (void *)u;
}

static void del_user_from_hash_table(User *u)
{
	User *tmp, *prev = NULL;

	for (tmp = userlist[u->hash]; tmp; tmp = tmp->next) {
		if (tmp == u) {
			if (prev)
				prev->next = tmp->next;
			else
				userlist[u->hash] = tmp->next;
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static void add_server_to_hash_table(char *name, Server *s)
{
	s->hash = HASH((unsigned char *) name, S_TABLE_SIZE);
	s->next = serverlist[s->hash];
	serverlist[s->hash] = (void *)s;
	if (IsDebug()) {
	        log("add_server_to_hash_table: %s, %d", name, s->hash);
	}
}

static void del_server_from_hash_table(Server *s)
{
	Server *tmp, *prev = NULL;

	for (tmp = serverlist[s->hash]; tmp; tmp = tmp->next) {
		if (tmp == s) {
			if (prev)
				prev->next = tmp->next;
			else
				serverlist[s->hash] = tmp->next;
			tmp->next = NULL;
			return;
		}
		prev = tmp;
	}
}

static User *new_user(char *nick)
{
	User *u;

	u = smalloc(sizeof(User));
	if (!nick)
		nick = "";
	memcpy(u->nick, nick, MAXNICK);
	add_user_to_hash_table(nick, u);

	return u;
}

#ifdef IRCD_2_10
void AddUser(char *nick, char *user, char *host, Server *server)
#else
void AddUser(char *nick, char *user, char *host, char *server)
#endif
{
	User *u;

	if (IsDebug()) {
#ifdef IRCD_2_10
		log("AddUser(): %s (%s@%s) -> %s", nick, user, host,
			server->name);
#else
		log("AddUser(): %s (%s@%s) -> %s", nick, user, host, server);
#endif
	}
	u = finduser(nick);
	if (u) {
		log("trying to add a user that already exists? (%s)", nick);
		return;
	}

	u = new_user(nick);
	u->hostname = sstrdup(host);
	u->username = sstrdup(user);
#ifdef IRCD_2_10
	u->server = server;
#else
	u->server = findserver(server);
#endif
	u->flags &= ~UFLAG_ISOPER;
	u->t_flood = time(NULL);
	u->flood = 0;
	u->flags &= ~UFLAG_ISAWAY;
#ifdef OPER_MODES
	u->operflags = 0;
#endif
	u->myuser = NULL;
	u->tld = AddTLD(u);

	if (!u->server)
		log("Unable to find server for user %s (%s?)", nick, server);
	else
		IncreaseUsers(u->server);
}

void DelUser(char *nick)
{
	User *u = finduser(nick);

	if (IsDebug()) {
		log("DelUser(%s)", nick);
	}

	if (!u) {
		log("DelUser(%s) failed!", nick);
		return;
	}

	del_user_from_hash_table(u);

	DecreaseUsers(u->server);

	if (u->tld)
		u->tld->users--;
	if (IsUIsAway())
		me.away--;
	if (IsUIsOper()) {
		me.opers--;
		u->server->opers--;
	}
	if (u->myuser)
		u->flags &= ~MFLAG_ISON;

	/* Debug asserts */
	assert(u->tld->users >= 0);
	assert(u->server->users >= 0);
	assert(u->server->opers >= 0);
	assert(me.opers >= 0);
	assert(me.away >= 0);
	assert(me.users >= 0);

	free(u->hostname);
	free(u->username);
	free(u);
}

void Change_User(User *u, char *newnick)
{
	if (IsDebug()) {
		log("Change_User(%s, %s)", u->nick, newnick);
	}

	del_user_from_hash_table(u);

	u->nick[1] = 0;
	memcpy(u->nick, newnick, MAXNICK);

	add_user_to_hash_table(u->nick, u);
}

User *finduser(char *nick)
{
	User *u;

	if (IsDebug()) {
		log("finduser(%s)", nick);
	}

	u = userlist[HASH((unsigned char *) nick, U_TABLE_SIZE)];
	while (u && (strcasecmp(u->nick, nick) != 0))
		u = u->next;

	if (IsDebug()) {
		log("finduser(%s) -> %s", nick, (u) ? u->nick : "NOTFOUND");
	}

	return u;
}

void init_user_hash()
{
	int i;
	User *u, *prev;

	for (i = 0; i < U_TABLE_SIZE; i++) {
		u = userlist[i];
		while (u) {
			prev = u->next;
	if (IsDebug()) {
		log("init_user_hash() deleting %s", u->nick);
	}
			free(u->hostname);
			free(u->username);
			free(u);

			u = prev;
		}
		userlist[i] = NULL;
	}
	bzero((char *)userlist, sizeof(userlist));
}

void UserMode(char *nick, char *modes)
{
	User *u = finduser(nick);
	int add = 0;

	if (!u) {
		log("Changing modes for unknown user: %s", nick);
		return;
	}

	/* while (*modes++) { */
	while (*modes) {
		switch(*modes) {
			case '+': add = 1;	break;
			case '-': add = 0;	break;
			case 'o':
				if (add) {
				    if (!IsUIsOper()) {
					u->flags |= UFLAG_ISOPER;
					IncreaseOpers(u->server);
#ifdef OPER_MODES
					u->operflags |= OPER_GLOBAL;
#endif
					if (IsDebug()) {
					    log("%s is now an IRCop.",
						u->nick);
					}
				    }
				} else {
				    if (IsUIsOper()) {
					u->flags &= ~UFLAG_ISOPER;
					me.opers--;
					u->server->opers--;
#ifdef OPER_MODES
					u->operflags = 0;
#endif
					if (IsDebug()) {
					    log("%s is no longer an IRCop.",
						u->nick);
					}
					assert(me.opers >= 0);
					assert(u->server->opers >= 0);
				    }
				}
				break;
#ifdef IRCD_2_10
			/* IRCnet propegates AWAY by setting usermode +a */
			case 'a':
				if (add) {
				    if (!IsUIsAway()) {
					u->flags |= UFLAG_ISAWAY;
					me.away++;
				    }
				} else {
				    if (IsUIsAway()) {
					u->flags &= ~UFLAG_ISAWAY;
					me.away--;
				    }
				    assert(me.away >= 0);
				}
				break;
#endif
#ifdef OPER_MODES
			case 'a':
				if (add)
					u->operflags |= OPER_SVSADMIN;
				else
					u->operflags &= ~OPER_SVSADMIN;
				break;
                        case 'N':
                                if (add)
                                        u->operflags |= OPER_NETADMIN;
                                else
                                        u->operflags &= ~OPER_NETADMIN;
                                break;
                        case 'C':
                                if (add)
                                        u->operflags |= OPER_COADMIN;
                                else
                                        u->operflags &= ~OPER_COADMIN;
                                break;
                        case 'T':
                                if (add)
                                        u->operflags |= OPER_TECHADMIN;
                                else
                                        u->operflags &= ~OPER_TECHADMIN;
                                break;
                        case 'A':
                                if (add)
                                        u->operflags |= OPER_ADMIN;
                                else
                                        u->operflags &= ~OPER_ADMIN;
                                break;
			case 'O':
				if (add)
					u->operflags |= OPER_LOCAL;
				else
					u->operflags &= ~OPER_LOCAL;
				break;
#endif /* OPER_MODES */
			default:
				break;
		}
		modes++;
	}
}

static Server *new_server(char *name, char *desc)
{
	Server *s;

	s = calloc(sizeof(Server), 1);
        memset(s, 0, sizeof(Server));
	if (!name)
		name = "";
	memcpy(s->name, name, MAXHOST);
	if (!desc)
		desc = "";
	memcpy(s->desc, desc, LINE);
	add_server_to_hash_table(name, s);

	if (IsDebug()) {
		log("new_server: %s, %s", name, desc);
	}

	return s;
}

#ifdef IRCD_2_10
void AddServer(char *name, int hops, int token, char *desc, char* uplink)
#else
void AddServer(char *name, int hops, char *desc, char* uplink)
#endif
{
	Server *s;

#ifndef FNMATCH
	if (strchr(name, '*')) {
		globops(me.name, "FATAL ERROR: FNMATCH must be #define'ed to process server: %s", name);
		log("FATAL ERROR: FNMATCH must be #define'ed to process server: %s", name);
		SendQuit(finduser(s_StatServ), 0, "FATAL ERROR");
		sleep(1);
		close(servsock);
		SaveStats();
		exit(0);
	}
#endif

	if (IsDebug()) {
		log("New Server: %s", name);
	}
	strlower(name);

	/* Check to see if a server already exists - Why is a mystery to me 
	** but sometimes on Hyb7 it'll happen
	*/
	s = findserver(name);
	if (s) {
		log("Server %s already exists in the server table", name);
	} else {
		s = new_server(name, desc);
	}
	s->hops = hops;
#ifdef IRCD_2_10
	s->token = token;
#endif
	s->connected_since = time(NULL);
	s->last_announce = time(NULL);
#ifdef LAG_CHECK
	s->ping = -1;
	s->lowest_ping = 0;
	s->highest_ping = 0;
	s->t_lowest_ping = time(NULL);
	s->t_highest_ping = time(NULL);
#endif /* LAG_CHECK */
        if (uplink == NULL)
                s->uplink = NULL;
        else
                s->uplink = findserver(uplink);
#ifdef TRACK_VERSION
	/* If it is NOT me.name */
	if (strcasecmp(me.name, name)) {
		sts(":%s VERSION %s", s_StatServ, name);
		s->version[0] = '\0';
	} else {
		snprintf(s->version, BUFSIZE, "%s(%s) %s %s -> %s %s",
			PATCHLEVEL, SERIALNUM, me.name, VersionFlags(),
			version_date, version_time);
	}
#endif
	IncreaseServers();
	AddStats(s);
}

/* This is similar to DelServer, but always recurses down the tree!
** This function is used when the connection to the uplink is lost
*/
void FlushServer(char *name)
{
        Server *up = NULL;
        User* u = NULL;
	Server *s = findserver(name);
        int i;
	if (IsDebug()) {
		log("FlushServer called for %s", name);
	}
        for (i=0; i<U_TABLE_SIZE; i++) {
                u=userlist[i];
                if (u) {
                        if (!strcasecmp(name, u->server->name)) {
                                DelUser(u->nick);
                        }
                }
        }
        for (i=0; i<S_TABLE_SIZE; i++) {
                up=serverlist[i];
                if (up) {
                        if (up->uplink) {
                                if (!strcasecmp(name, up->uplink->name)) {
                                        FlushServer(up->name);
                                }
                        }
                }
        }

        if (!s || !name) {
                log("DelServer(): %s failed!", name);
                return;
        }

        del_server_from_hash_table(s);
        s->stats->lastseen = time(NULL);
        DecreaseServers();
        free(s);
}

void DelServer(char *name)
{
	Server *s = findserver(name);

#if defined(IRCD_HYB6) || defined(IRCD_HYB7) || defined(IRCD_2_10) || defined(IRCD_DALNQ) || defined(IRCD_OPN)
	/* This is to support the QS capab
	** IRCnet 2.10.x does this too
	** Dalnet servers supporting the NOQUIT PROTOCTL also are 
	** supported in here
	*/
	Server *up = NULL;
	User* u = NULL;
	int i;
	for (i=0; i<U_TABLE_SIZE; i++) {
		u=userlist[i];
		if (u) {
			if (!strcmp(name, u->server->name)) {
				DelUser(u->nick);
			}
		}
	}
	for (i=0; i<S_TABLE_SIZE; i++) {
		up=serverlist[i];
		if (up) {
			if (up->uplink) {
				if (!strcmp(name, up->uplink->name)) {
					DelServer(up->name);
				}
			}
		}
	}
#endif

	if (!s || !name) {
		log("DelServer(): %s failed!", name);
		return;
	}

	del_server_from_hash_table(s);
	s->stats->lastseen = time(NULL);
	DecreaseServers();
	assert(me.servers >= 0);
	free(s);
}

#ifdef IRCD_2_10
Server *findservtoken(int token)
{
	Server *s = NULL;

        while (s && (s->token != token))
                s = s->next;
        if (s)
                return s;

	/* We should never get here */
	if (IsDebug()) {
        	log("findservtoken(%d) -> %s", token,
			(s) ? s->name : "NOT FOUND");
	}

	/* I'd rather return an explicit NULL here */
        return NULL;
}
#endif

Server *findserver(char *name)
{
	Server *s;
#ifdef FNMATCH
	int i;
#endif

	strlower(name); 
	if (IsDebug()) {
		log("findserver(%s)", name);
	}
	s = serverlist[HASH((unsigned char *) name, S_TABLE_SIZE)];
	while (s && (strcasecmp(s->name, name) != 0))
		s = s->next;
	if (s)
		return s;

#ifdef FNMATCH
/* The name in the server table is the pattern (with the *) 
** The name is what is passed into this function (usually full 
** server name)
**
** NOTE:  This also is a performance hit, if there are a number of masked
**        servers or there are a lot of users on that masked server/set of
**        servers hidden by the mask
*/
	for(i=0; i<S_TABLE_SIZE; i++) {
		s = serverlist[i];
        	while (s && (fnmatch(s->name, name, 0) != 0))
                	s = s->next;
		if (s)
			return s;
	}
#endif /* FNMATCH */

/* We should never get here unless the server doesn't exist */
	if (IsDebug()) {
		log("findserver(%s) -> %s", name, (s) ? s->name : "NOT FOUND");
	}

/* I'd rather return an explicit NULL here */
	return NULL;
}

void init_server_hash()
{
	int i;
	Server *s, *b;

	for (i = 0; i < S_TABLE_SIZE; i++) {
		s = serverlist[i];
		while (s) {
			b = s->next;
	if (IsDebug()) {
		log("init_server_hash() deleting %s", s->name);
	}
			free(s);
			s = b;
		}
		serverlist[i] = NULL;
	}
	bzero((char *)serverlist, sizeof(serverlist));
#ifdef IRCD_2_10
	AddServer(me.name, 0, 0, me.desc, NULL);
#else
	AddServer(me.name, 0, me.desc, NULL);
#endif
}

void LoadMyUsers()
{
	FILE *fp;
	char buf[BUFSIZE];
	char *user, *pass, *timestamp;
	MyUser *myu;

	if ((fp = fopen(USERFILE, "r")) == NULL) {
		fprintf(stderr, "Please type this: echo \"<nickname> <password> 0 >>"
		"%s\"\n", USERFILE);
		fprintf(stderr, "Replace nickname with your nickname, and password"
		" with your password, encrypting if necessary\n");
		fprintf(stderr, "This is used for the LOGIN feature of Stats.\n");
		log("Unable to open userlist.");
		exit(0);
		myuhead = NULL;
	}

	while (fgets(buf, BUFSIZE, fp)) {
		user = strtok(buf, " ");
		pass = strtok(NULL, " ");
		timestamp = strtok(NULL, "\n");

		if ((!user) || (!pass)) {
			fprintf(stderr, "Invalid entry in user file: %s\n",
				USERFILE);
			fprintf(stderr, "Format: nick pass 0\n");
			fprintf(stderr, "Or: nick pass <time last seen>\n");
			fprintf(stderr, "Exiting...\n");
			exit(0);
		}

                myu = smalloc(sizeof(MyUser));
                myu->username = strdup(user);
                myu->password = strdup(pass);
		myu->flags &= ~MFLAG_ISON;

		if (!timestamp) {
			log("Invalid time value in user file %s, defaulting to 0", USERFILE);
			myu->lastseen = 0;
		} else {
			myu->lastseen = atol(timestamp);
		}

		if (!myuhead) {
			myuhead = myu;
			myuhead->next = NULL;
		} else {
			myu->next = myuhead;
			myuhead = myu;
		}
	}
	fclose(fp);
}

void SaveMyUsers()
{
	MyUser *myu;
	FILE *fp;

	if (!myuhead)
		return;

	if ((fp = fopen(USERFILE, "w")) == NULL) {
		log("Unable to open %s for writing.", USERFILE);
		return;
	}
	for (myu = myuhead; myu; myu = myu->next) {
		fprintf(fp, "%s %s %ld\n", myu->username, myu->password,
			myu->lastseen);
	}
	fclose(fp);
}

void DeleteMyUser(char *username)
{
	MyUser *myu, *prev = NULL;

	if (!myuhead)
		return;

	myu = myuhead;

	while (myu) {
		if (!strcasecmp(myu->username, username))
			break;
		prev = myu;
		myu = myu->next;
	}
	if (!myu)
		return;

	if (!prev) {
		myuhead = myu->next;
	} else if (myu->next == NULL) {
		prev->next = NULL;
	} else {
		prev->next = myu->next;
	}
	free(myu->username);
	free(myu->password);
	free(myu);
}

MyUser *findmyuser(char *nick)
{
	MyUser *myu;

	for (myu = myuhead; myu; myu = myu->next) {
		if (!strcasecmp(nick, myu->username))
			return myu;
	}
	return NULL;
}
